import requests
from cachetools import TTLCache
from omnik.ha_logger import hybridlogger
from decimal import Decimal

# TODO: Create Abstract Base Class


class BasePlugin(type):
    def __init__(cls, name, bases, attrs):
        super(BasePlugin, cls).__init__(name)
        if not hasattr(cls, "plugins"):
            cls.plugins = []
        else:
            cls.register(cls)  # Called when a plugin class is imported

    def register(cls, plugin=None):
        cls.plugins.append(plugin())


class Plugin(object, metaclass=BasePlugin):

    config = None
    logger = None
    hass_api = None
    process_aggregates = False

    cache = TTLCache(maxsize=1, ttl=300)

    def jsonval(self, value):
        if isinstance(value, Decimal):
            if float(value) == int(value):
                return int(value)
            else:
                return float(value)
        elif isinstance(value, float):
            return value
        elif isinstance(value, int):
            return value
        elif isinstance(value, bool):
            return value
        elif isinstance(value, str):
            return value
        else:
            return str(value)

    def get_weather(self):
        try:
            if "weather" not in self.cache:
                self.logger.debug("[cache miss] Fetching weather data")
                url = "https://{endpoint}/data/2.5/weather?lon={lon}&lat={lat}&units={units}&APPID={api_key}".format(
                    endpoint=self.config.get(
                        "openweathermap", "endpoint", fallback="api.openweathermap.org"
                    ),
                    lat=self.config.get("openweathermap", "lat"),
                    lon=self.config.get("openweathermap", "lon"),
                    units=self.config.get("openweathermap", "units", fallback="metric"),
                    api_key=self.config.get("openweathermap", "api_key"),
                )

                res = requests.get(url)

                res.raise_for_status()

                self.cache["weather"] = res.json()

            return self.cache["weather"]

        except requests.exceptions.HTTPError as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                "Unable to get weather data. [{0}]: {1}".format(
                    type(e).__name__, str(e)
                ),
            )
            raise e

    def process(self, **args):
        pass

    def terminate(self):
        pass



import hashlib
from requests import Request, Session
from omnik.ha_logger import hybridlogger
from omnik.plugin_client import Client
from decimal import Decimal


DATAMAP = {
    "DV1": "voltage_pv1",
    "DV2": "voltage_pv2",
    "DC1": "current_pv1",
    "DC2": "current_pv2",
    "AV1": "voltage_ac1",
    "AV2": "voltage_ac2",
    "AV3": "voltage_ac3",
    "AC1": "current_ac1",
    "AC2": "current_ac2",
    "AC3": "current_ac3",
    "APo_t1": "current_power",
    "A_Fo1": "frequency_ac1",
    "A_Fo2": "frequency_ac2",
    "A_Fo3": "frequency_ac3",
    "Et_ge0": "total_energy",
    "Etdy_ge1": "today_energy",
    "INV_T0": "inverter_temperature",
    "t_w_hou1": "operation_hours",
    "INV_ST1": "inverter_status",
    "PG_V_ERR0": "grid_voltage_error_value",  # Default "0.0"
    "PG_F_ERR0": "grid_frequency_error_value",  # Default "0.0"
    "PG_I_ERR1": "grid_impedance_error_value",  # Default "65.535"
    "MAC_T_ERRin1": "inner_temperature_error value",  # Default "0.0"
    "V_ERRi1": "input_voltage_1_error_value",  # Default "0.0"
    "V_ERRi2": "input_voltage_2_error_value",  # Default "0.0"
    "V_ERRi3": "input_voltage_3_error_value",  # Default "0.0"
    "ELC_ERR1": "leak_current_error_value",  # Default "0.0"
}


class OmnikPortalClient(Client):
    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(
            self.logger, self.hass_api, "INFO", "Client enabled: solarmanpv client"
        )

        # API Key's
        self.app_id = self.config.get(
            "client.solarmanpv", "app_id", fallback="202107211350206"
        )
        self.app_key = self.config.get(
            "client.solarmanpv", "app_key", fallback="61510dabe03526bfcdfa94c9ed2560cc"
        )

        self.base_url = self.config.get(
            "client.solarmanpv", "base_url", fallback="https://api.solarmanpv.com"
        )

        self.username = self.config.get("client.solarmanpv", "username")
        self.password = self.config.get("client.solarmanpv", "password")

        self.session = Session()

        self.last_update_cache = {}

    def initialize(self):
        url = f"{self.base_url}/account/v1.0/token"

        body = {
            "email": self.username,
            "password": hashlib.sha256(str.encode(self.password)).hexdigest(),
            "appSecret": self.app_key,
        }

        if not self.app_id or not self.app_key:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                "Authentication error! Please setup app_id and app_key!",
            )
            return None

        data = self._api_request(url, body)

        if not data or not data.get("success"):
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                "Authorization failed!",
            )
            return None
        else:
            hybridlogger.ha_log(
                self.logger, self.hass_api, "DEBUG", "Account validation successful!"
            )

        # set authentication header
        self.session.headers.update({"Authorization": "Bearer " + data["access_token"]})
        return True

    def _api_request(self, url, json):
        headers = {
            "User-Agent": "Omnikdatalogger",
        }

        #        if self.token:
        #            headers["Authorization"] = "Bearer " + self.token

        params = {"appId": self.app_id}

        request = Request("POST", url=url, params=params, json=json, headers=headers)
        prepped_request = self.session.prepare_request(request)

        response = self.session.send(prepped_request)

        response.raise_for_status()

        return response.json()

    def getPlants(self):
        # Get station list
        url = f"{self.base_url}/station/v1.0/list"
        json = {}

        stationlist = self._api_request(url, json=json)
        if not stationlist or not stationlist.get("success"):
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                "plant/station list cannot be loaded from the cloud config, no valid data available",
            )
            return None

        # Get INVERTER devices for retreived stations
        data = []
        for station in stationlist.get("stationList"):
            # Only append stations that have a valid INVERTER device
            url = f"{self.base_url}/station/v1.0/device"
            json = {"stationId": station.get("id")}
            devicelist = self._api_request(url, json=json)
            for device in devicelist.get("deviceListItems"):
                if device.get("deviceType") == "INVERTER":
                    data.append(
                        {
                            "plant_id": f'{str(station.get("id"))},{str(device.get("deviceSn"))}',
                            "last_update": device.get("collectionTime"),
                        }
                    )

        return data

    def getPlantData(self, plant_id):

        # we collect our data in `data` but we need a serial number to query the API, we extract this from plant_id
        data = {}
        station, serial = plant_id.split(",")
        data["inverter"] = serial

        # get communication status and last update
        json = {"deviceSn": serial}
        url = f"{self.base_url}/device/v1.0/communication"

        communicationdata = self._api_request(url, json=json)
        if not communicationdata or not communicationdata.get("success"):
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"Plant/station communication data for {station}, {serial} cannot be loaded from the cloud config, no valid data available",
            )
            return None

        # update last update
        data["last_update"] = communicationdata["communication"].get("updateTime")

        # only update details if the updateTime has changed
        if (
            self.last_update_cache.get(serial)
            and self.last_update_cache.get(serial) == data["last_update"]
        ):
            # Nothing has changed no need for updates
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"Plant/station communication data for {station}, {serial} has no new update, ignoring. Is your update frequency to high?",
            )

        url = f"{self.base_url}/device/v1.0/currentData"

        devicedata = self._api_request(url, json=json)

        if not devicedata or not devicedata.get("success"):
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"Plant data for station ({station}) cannot be loaded, no valid data available",
            )
            return None

        if devicedata["deviceState"] == 3:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"plant data ({plant_id}) indicates inverter is offline",
            )
        elif devicedata["deviceState"] == 2:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"plant data ({plant_id}) indicates inverter state is alerting",
            )

        for entry in devicedata.get("dataList"):
            # Adjust power to Watt (not kW) and covert strings numbers to float or int
            if entry["key"] in DATAMAP:
                data[DATAMAP[entry["key"]]] = Decimal(entry["value"])

        # update cache
        self.last_update_cache[serial] = data["last_update"]

        return data



from omnik.ha_logger import hybridlogger
from omnik.plugin_localproxy import LocalProxyPlugin
import paho.mqtt.client as mqttclient
import uuid
import json
import binascii


class MQTTproxy(LocalProxyPlugin):

    """
    This plugin enables you to subscribe to the logger requests using MQTT
    Use scripts/proxy/omnikloggerproxy.py to capture the inverter events and publish the data to MQTT
    If you are using Home Assistant consider a setup with AppDaemon and the hassapi plugin to collect the published data

    This class makes use the following localproxy client objects
    self.client.msg
    self.client.msgevent
    self.client.semaphore
    self.client.inverters
    self.client.plant_id_list
    """

    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(
            self.logger, self.hass_api, "INFO", "localproxy client plugin: MQTTproxy"
        )
        self.logger_sensor_name = self.mqttconfig("logger_sensor_name", "Datalogger")
        self.mqtt_discovery_prefix = self.mqttconfig(
            "discovery_prefix", "homeassistant"
        )
        self.mqtt_host = self.mqttconfig("host", "localhost")
        self.mqtt_port = int(self.mqttconfig("port", "1883"))

        self.tls = bool(self.mqttconfig("tls", fallback=False))
        self.ca_certs = self.config.get("output.mqtt", "ca_certs", fallback=None)
        self.client_cert = self.config.get("output.mqtt", "client_cert", fallback=None)
        self.client_key = self.config.get("output.mqtt", "client_key", fallback=None)

        self.mqtt_client_name_prefix = self.mqttconfig(
            "client_name_prefix", "ha-mqttproxy-omniklogger"
        )
        self.mqtt_client_name = self.mqtt_client_name_prefix + "_" + uuid.uuid4().hex
        self.mqtt_username = self.mqttconfig("username", None)
        self.mqtt_password = self.mqttconfig("password", None)
        if not self.mqtt_username or not self.mqtt_password:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                "Please specify MQTT username and password in the configuration",
            )
        # mqtt setup
        self.mqtt_client = mqttclient.Client(self.mqtt_client_name)
        self.mqtt_client.on_connect = self._mqtt_on_connect  # bind call back function
        self.mqtt_client.on_disconnect = (
            self._mqtt_on_disconnect
        )  # bind call back function
        self.mqtt_client.on_message = (
            self._mqtt_on_message
        )  # called on receiving updates on subscibed messages
        self.mqtt_client.username_pw_set(self.mqtt_username, self.mqtt_password)
        # TLS support
        if self.tls:
            self.mqtt_client.tls_set(
                ca_certs=self.ca_certs or None,
                certfile=self.client_cert or None,
                keyfile=self.client_key or None,
            )

    def _mqtt_on_connect(self, client, userdata, flags, rc=0, properties=None):
        if rc == 0:
            hybridlogger.ha_log(
                self.logger, self.hass_api, "INFO", "MQTTproxy client connected"
            )
            # subscribe to datalogger objects
            for inverter in self.client.inverters:
                topic = (
                    f"{self.mqtt_discovery_prefix}/binary_sensor/{self.logger_sensor_name}_"
                    f"{self.client.inverters[inverter]['inverter_sn']}/attr"
                )
                try:
                    self.mqtt_client.subscribe(topic)
                    hybridlogger.ha_log(
                        self.logger,
                        self.hass_api,
                        "INFO",
                        f"MQTTproxy subscribed on topic: {topic}",
                    )
                except Exception as e:
                    hybridlogger.ha_log(
                        self.logger,
                        self.hass_api,
                        "ERROR",
                        f"MQTTproxy failed to subsrcibed ontopic: {topic}. Error {e}",
                    )

    def _mqtt_on_disconnect(self, client, userdata, flags, rc=0):
        if rc == 0:
            hybridlogger.ha_log(self.logger, self.hass_api, "INFO", "MQTT disconnected")

    def _mqtt_on_message(self, client, userdata, message):
        # Try to parse payload as json
        try:
            payload = json.loads(message.payload)
            data = binascii.a2b_base64(payload["data"])
            if len(data) == 128:
                # Fill data structure
                self.client.semaphore.acquire()
                self.client.msg["data"] = data
                self.client.msg["isSet"] = True
                self.client.msg["plugin"] = __name__
                self.client.semaphore.release()
                # Trigger processing the message
                self.client.msgevent.set()
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"MQTTproxy: invalid data received on topic: {message.topic}. Error {e}",
            )

    def mqttconfig(self, parameter, fallback):
        return self.config.get(
            "client.localproxy.mqtt_proxy",
            parameter,
            self.config.get("output.mqtt", parameter, fallback=fallback),
        )

    def terminate(self):
        self.mqtt_client.loop_stop()
        self.mqtt_client.disconnect()

    def listen(self):
        self.mqtt_client.connect(self.mqtt_host, self.mqtt_port)
        # start processing messages
        self.mqtt_client.loop_start()



from datetime import datetime, timedelta
import pytz

try:
    from astral import sun
    from astral.geocoder import database, lookup

    VERSION = 2
except ImportError:
    from astral import Astral

    VERSION = 1

global default_city_name

default_city_name = "Amsterdam"


class daylight(object):
    def __init__(self, city_name=default_city_name):
        if VERSION == 2:
            self._city = lookup("Amsterdam", database())
        else:
            self._a = Astral()
            self._a.solar_depression = "civil"
            self._city = self._a[city_name]
            self._timezone = pytz.timezone(self._city.timezone)

    def sun(self, t=None):
        if not t:
            t = self.localtime()
        if VERSION == 2:
            return sun.sun(self._city.observer, t, tzinfo=self._city.timezone)
        else:
            return self._city.sun(t)

    def localtime(self):
        return datetime.utcnow().astimezone(pytz.timezone(self._city.timezone))

    @property
    def version(self):
        return VERSION

    @property
    def dawn(self):
        return self.sun()["dawn"]

    @property
    def next_dawn(self):
        _sundawn = self.dawn
        if _sundawn < self.localtime():
            _sundawn = self.sun(self.localtime() + timedelta(days=1))["dawn"]
        return _sundawn

    @property
    def sunrise(self):
        return self.sun()["sunrise"]

    @property
    def noon(self):
        return self.sun()["noon"]

    @property
    def sunset(self):
        return self.sun()["sunset"]

    @property
    def dusk(self):
        return self.sun()["dusk"]

    def sun_rising(self, time: datetime = None):
        if not time:
            time = self.localtime()
        _sun = self.sun(time)
        return (time > _sun["dawn"]) and (time < _sun["sunrise"])

    def sun_up(self, time: datetime = None):
        if not time:
            time = self.localtime()
        _sun = self.sun(time)
        return (time > _sun["sunrise"]) and (time < _sun["sunset"])

    def sun_shine(self, time: datetime = None):
        if not time:
            time = self.localtime()
        _sun = self.sun(time)
        return (time > _sun["dawn"]) and (time < _sun["dusk"])

    def sun_setting(self, time: datetime = None):
        if not time:
            time = self.localtime()
        _sun = self.sun(time)
        return (time > _sun["sunset"]) and (time < _sun["dusk"])

    def sun_down(self, time: datetime = None):
        if not time:
            time = self.localtime()
        _sun = self.sun(time)
        return (time < _sun["dawn"]) or (time > _sun["dusk"])


def main():
    dl = daylight("Amsterdam")
    nu = dl.localtime()
    print("Version: %s" % str(dl.version))
    print("Now:     %s" % str(nu))
    print("Dawn:    %s" % str(dl.dawn))
    print("nxtdawn: %s" % str(dl.next_dawn))
    print("Sunrise: %s" % str(dl.sunrise))
    print("Noon:    %s" % str(dl.noon))
    print("Sunset:  %s" % str(dl.sunset))
    print("Dusk:    %s" % str(dl.dusk))
    print("dDawn:    %s" % str(dl.dawn - nu))
    print("dSunrise: %s" % str(dl.sunrise - nu))
    print("dNoon:    %s" % str(dl.noon - nu))
    print("dSunset:  %s" % str(dl.sunset - nu))
    print("dDusk:    %s" % str(dl.dusk - nu))
    print("Sun rising :", dl.sun_rising())
    print("Sun up     :", dl.sun_up())
    print("Sun shine  :", dl.sun_shine())
    print("Sun setting:", dl.sun_setting())
    print("Sun down   :", dl.sun_down())


if __name__ == "__main__":
    main()



from omnik.ha_logger import hybridlogger
import omnik.InverterMsg
from omnik.plugin_client import Client
from socket import socket, error as sockerror, AF_INET, SOCK_STREAM
import binascii
import re
from datetime import datetime
import time
from decimal import Decimal
from requests import Request, Session


class TCPclient(Client):

    # IMPORTANT : ONLY WIFI Modules with s/n 602xxxxxx and 604xxxxxx support direct access thru port 8899!
    # For certain models it is possible to readout http:/{inverter_ip}:80/js/status.js If access through 8899 fails this method will be used as fall back method.
    # Based on: https://github.com/woutrrr/omnik-data-logger by Wouter van der Zwan
    # The script was tested only using simulations, let me know if this works for you

    def __init__(self):
        super().__init__()
        hybridlogger.ha_log(
            self.logger, self.hass_api, "INFO", "Client enabled: TCPclient"
        )

        # Get plant_id_list
        self.plant_id_list = self.config.getlist(
            "client.tcpclient", "plant_id_list", fallback=[]
        )
        if not self.plant_id_list:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "ERROR",
                "plant_id_list was not specified for client TCPclient",
            )
            raise Exception("plant_id_list was not specified")
        # Initialize dict with inverter info
        self.inverters = {}
        self.session = {}

        # self._mode indicates the access mode that will be used
        # 0: Not initialized
        # 1: Running in native mode (port 8899)
        # 2: Running in fallback mode (port 80) fetching http://{inverter_ip}:80/js/status.js
        self._mode = 0

    def getPlants(self):
        data = []
        inverterdata = {}
        for plant in self.plant_id_list:
            inverter_address = self.config.get(
                f"plant.{plant}", "inverter_address", fallback=None
            )
            if not inverter_address:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "ERROR",
                    f"inverter_address for plant {plant} was not specified [TCPclient]",
                )
                raise Exception("an inverter_address was not specified")
            inverter_port = int(
                self.config.get(f"plant.{plant}", "inverter_port", fallback="8899")
            )
            inverter_connection = (inverter_address, int(inverter_port))
            logger_sn = self.config.get(f"plant.{plant}", "logger_sn", fallback=None)
            if not logger_sn:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "ERROR",
                    "logger_sn (The serial number of the "
                    "Wi-Fi datalogger) for plant {plant} was not specified for [TCPclient], http mode only",
                )
            inverter_sn = self.config.get(
                f"plant.{plant}", "inverter_sn", fallback=None
            )
            if not inverter_sn:
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "WARNING",
                    "inverter_sn (The serial number of the inverter) "
                    "for plant {plant} was not specified for [TCPclient], http mode only",
                )
            inverterdata = {
                "inverter_address": inverter_address,
                "inverter_port": inverter_port,
                "logger_sn": int(logger_sn),
                "inverter_sn": inverter_sn,
                "inverter_connection": inverter_connection,
            }
            self.inverters[plant] = inverterdata
            data.append({"plant_id": plant})

        hybridlogger.ha_log(
            self.logger, self.hass_api, "DEBUG", f"plant list from config {data}"
        )

        return data

    def getPlantData(self, plant_id):
        data = None
        if not self._mode:
            # native mode
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Initializing: Trying to reach the inverter for plant {plant_id} over port 8899.",
            )
        if (
            self._mode <= 1
            and self.inverters[plant_id].get("inverter_sn")
            and self.inverters[plant_id].get("logger_sn")
        ):
            data = self._getPlantData_native(plant_id)
        if data:
            # set mode to native
            self._mode = 1
        if not self._mode:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Initializing: Trying to reach the inverter for plant {plant_id} over port http (fallback).",
            )
        if self._mode == 0 or self._mode == 2:
            # fall back mode
            data = self._getPlantData_fallback(plant_id)

        return data

    def _getPlantData_native(self, plant_id):
        # Create a TCP/IP socket
        valid = False
        data = {}
        # Create request message
        requestmsg = omnik.InverterMsg.request_string(
            self.inverters[plant_id]["logger_sn"]
        )
        sock = socket(AF_INET, SOCK_STREAM)
        sock.settimeout(40)
        try:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "DEBUG",
                "Connecting to %s port %s"
                % self.inverters[plant_id]["inverter_connection"],
            )
            sock.connect(self.inverters[plant_id]["inverter_connection"])
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "DEBUG",
                "Sending request message with logger_sn: "
                f"{omnik.InverterMsg.request_string(self.inverters[plant_id]['logger_sn'])}",
            )
            # Sending request message
            sock.sendall(requestmsg)
            # Wait for response (listen and block thread)
            rawmsg = sock.recv(129)
            # Validate the data response. Check length and serial number
            if len(rawmsg) >= 99:
                inverterMsg = omnik.InverterMsg.InverterMsg(rawmsg)
                serialnr = inverterMsg.getID()
                # lookup plant_id to see it is in the plants list
                if self.inverters[plant_id]["inverter_sn"] == serialnr:
                    data["plant_id"] = plant_id
                    valid = True
                hybridlogger.ha_log(
                    self.logger,
                    self.hass_api,
                    "DEBUG",
                    f"data size: {len(rawmsg)}.\n"
                    f"Datadump: {str(binascii.b2a_base64(rawmsg))}",
                )
        except sockerror as warn:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"Inverter is not reachable over port 8899, this is normal when it is dark. {warn}",
            )
            return None

        if valid:
            # Get the data from the received message
            inverterMsg.FetchDataDict(data)
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "INFO",
                f"New message received from inverter '{serialnr}'",
            )
        else:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                "Invalid response, ignoring message",
            )
            data = None

        return data

    def _getPlantData_fallback(self, plant_id):
        data = None
        # Get get url for debugging
        url = self.config.get(
            f"plant.{plant_id}",
            "url",
            fallback=f"http://{self.inverters[plant_id].get('inverter_address')}/js/status.js",
        )
        headers = {
            "User-Agent": "Omnikdatalogger",
            "Content-Type": "text/xml; charset=utf-8",
            "Connection": "keep-alive",
            "Accept": "Content-Type: text/html; charset=UTF-8",
        }
        if not self.session.get(plant_id):
            self.session[plant_id] = Session()
        request = Request("GET", url=url, headers=headers)
        prepped_request = self.session[plant_id].prepare_request(request)
        response = self.session[plant_id].send(prepped_request)
        # Throw if status is not valid
        response.raise_for_status()

        # We have response, lets validate it now
        inverter_data_search = re.search(
            r'(?<=webData=").*?(?=";)|(?<=myDeviceArray\[0\]=").*?(?=";)',
            response.content.decode("utf-8"),
        )
        if inverter_data_search:
            # Now extract our data (if valid)
            inverter_data = inverter_data_search.group(0).split(",")
            lastupdate = time.time()
            data = {
                "plant_id": plant_id,
                "last_update": lastupdate,
                "last_update_time": datetime.utcfromtimestamp(lastupdate).strftime(
                    "%Y-%m-%dT%H:%M:%SZ"
                ),
                "inverter": inverter_data[0],
                "current_power": Decimal(inverter_data[5]),
                "today_energy": Decimal(inverter_data[6]) / 100,
                "total_energy": Decimal(inverter_data[7]) / 10,
            }
            # set mode to http
            self._mode = 2
        return data



from dsmr_parser import obis_references
from .terminal import Terminal
from decimal import Decimal
import threading
from datetime import datetime
from omnik.ha_logger import hybridlogger


class DSRM(object):

    config = None
    logger = None
    hass_api = None
    datalogger = None

    def initialize(self, terminals, dsmr_callback):
        # args.host, args.port, args.version, args.device
        self.dsmr_callback = dsmr_callback
        # self.stop = False

        self._init_terminals(terminals)

    def _init_terminals(self, terminals):
        self.terminals = {}
        self.sync = {}
        self.cache = {}
        self.ts_last_telegram = {}
        self.last_gas_update = {}
        self.tconfig = {}
        self.tconfig["tarif"] = {}
        if self.config.has_option("dsmr", "tarif"):
            tarif_list = self.config.getlist("dsmr", "tarif")
        else:
            tarif_list = ["0001", "0002"]
            self.tconfig["tarif"]["0001"] = "low"
            self.tconfig["tarif"]["0002"] = "normal"
        for tarif in tarif_list:
            if tarif in self.tconfig["tarif"]:
                default = self.tconfig["tarif"][tarif]
            else:
                default = tarif
            self.tconfig["tarif"][tarif] = self.config.get(
                "dsmr", f"tarif.{tarif}", default
            )
        for terminal in terminals:
            # Pre read configuration for each terminal
            self.tconfig[terminal] = {}

            self.tconfig[terminal]["plant_id"] = self.config.get(
                f"dsmr.{terminal}", "plant_id", "0"
            )
            self.tconfig[terminal]["gas_meter"] = self.config.getboolean(
                f"dsmr.{terminal}", "gas_meter", True
            )
            self.tconfig[terminal]["dsmr_version"] = self.config.get(
                f"dsmr.{terminal}", "dsmr_version ", "5"
            )
            self.tconfig[terminal]["total_energy_offset"] = Decimal(
                self.config.get(f"dsmr.{terminal}", "total_energy_offset", "0")
            )

            # Init terminal sync parameters
            self.sync[terminal] = 0
            self.cache[terminal] = Decimal(1000000)
            self.ts_last_telegram[terminal] = 0
            self.last_gas_update[terminal] = [0, Decimal("0.0"), Decimal("0.000")]

            # Initialize terminal
            self.terminals[terminal] = Terminal(
                self.config,
                self.logger,
                self.hass_api,
                terminal,
                self.dsmr_serial_callback,
                self.tconfig[terminal]["dsmr_version"],
            )

    def terminate(self):
        # cleanup connection after user initiated shutdown
        for terminal in self.terminals:
            self.terminals[terminal].terminate()

    def _proces_power_current_fase(self, fase, msg_dsmr, telegram):

        # Build reference arrays
        voltage_fase_obis = [
            obis_references.INSTANTANEOUS_VOLTAGE_L1,
            obis_references.INSTANTANEOUS_VOLTAGE_L2,
            obis_references.INSTANTANEOUS_VOLTAGE_L3,
        ]
        voltage_fase_sag_obis = [
            obis_references.VOLTAGE_SAG_L1_COUNT,
            obis_references.VOLTAGE_SAG_L2_COUNT,
            obis_references.VOLTAGE_SAG_L3_COUNT,
        ]
        voltage_fase_swell_obis = [
            obis_references.VOLTAGE_SAG_L1_COUNT,
            obis_references.VOLTAGE_SAG_L2_COUNT,
            obis_references.VOLTAGE_SAG_L3_COUNT,
        ]
        current_fase_obis = [
            obis_references.INSTANTANEOUS_CURRENT_L1,
            obis_references.INSTANTANEOUS_CURRENT_L2,
            obis_references.INSTANTANEOUS_CURRENT_L3,
        ]
        pwr_pos_fase_obis = [
            obis_references.INSTANTANEOUS_ACTIVE_POWER_L1_POSITIVE,
            obis_references.INSTANTANEOUS_ACTIVE_POWER_L2_POSITIVE,
            obis_references.INSTANTANEOUS_ACTIVE_POWER_L3_POSITIVE,
        ]
        pwr_neg_fase_obis = [
            obis_references.INSTANTANEOUS_ACTIVE_POWER_L1_NEGATIVE,
            obis_references.INSTANTANEOUS_ACTIVE_POWER_L2_NEGATIVE,
            obis_references.INSTANTANEOUS_ACTIVE_POWER_L3_NEGATIVE,
        ]

        sfase = f"{fase}"

        if pwr_pos_fase_obis[fase - 1] in telegram:
            msg_dsmr[f"INSTANTANEOUS_CURRENT_L{sfase}"] = telegram[
                current_fase_obis[fase - 1]
            ].value
            msg_dsmr[f"VOLTAGE_SAG_L{sfase}_COUNT"] = telegram[
                voltage_fase_sag_obis[fase - 1]
            ].value
            msg_dsmr[f"VOLTAGE_SWELL_L{sfase}_COUNT"] = telegram[
                voltage_fase_swell_obis[fase - 1]
            ].value
            msg_dsmr[f"INSTANTANEOUS_ACTIVE_POWER_L{fase}_POSITIVE"] = telegram[
                pwr_pos_fase_obis[fase - 1]
            ].value
            msg_dsmr[f"INSTANTANEOUS_ACTIVE_POWER_L{fase}_NEGATIVE"] = telegram[
                pwr_neg_fase_obis[fase - 1]
            ].value
            # current_net_power_l{fase} is calculated in Watt and is negative when power is delivered
            msg_dsmr[f"current_net_power_l{sfase}"] = (
                (
                    msg_dsmr[f"INSTANTANEOUS_ACTIVE_POWER_L{sfase}_POSITIVE"]
                    - msg_dsmr[f"INSTANTANEOUS_ACTIVE_POWER_L{sfase}_NEGATIVE"]
                )
                * Decimal("1000")
            ).quantize(Decimal("1."))
        if voltage_fase_obis[fase - 1] in telegram:
            # We have voltage, so lets calculate the current using power and voltage
            msg_dsmr[f"INSTANTANEOUS_VOLTAGE_L{sfase}"] = telegram[
                voltage_fase_obis[fase - 1]
            ].value
            # current_net_current_l{fase} is calculated in Ampère and is negative when power is delivered
            msg_dsmr[f"net_current_l{sfase}"] = (
                msg_dsmr[f"current_net_power_l{sfase}"]
                / msg_dsmr[f"INSTANTANEOUS_VOLTAGE_L{sfase}"]
            ).quantize(Decimal(".01"))

    def _process_power_details(self, msg_dsmr, telegram):
        terminal = threading.currentThread().getName()
        try:
            # Get global power (kWh)
            msg_dsmr["CURRENT_ELECTRICITY_USAGE"] = telegram[
                obis_references.CURRENT_ELECTRICITY_USAGE
            ].value
            msg_dsmr["CURRENT_ELECTRICITY_DELIVERY"] = telegram[
                obis_references.CURRENT_ELECTRICITY_DELIVERY
            ].value
            # Get energy usage
            msg_dsmr["ELECTRICITY_USED_TARIFF_1"] = telegram[
                obis_references.ELECTRICITY_USED_TARIFF_1
            ].value
            msg_dsmr["ELECTRICITY_USED_TARIFF_2"] = telegram[
                obis_references.ELECTRICITY_USED_TARIFF_2
            ].value
            msg_dsmr["energy_used_net"] = (
                msg_dsmr["ELECTRICITY_USED_TARIFF_1"]
                + msg_dsmr["ELECTRICITY_USED_TARIFF_2"]
            )  # ELECTRICITY_USED (NV)
            msg_dsmr["ELECTRICITY_DELIVERED_TARIFF_1"] = telegram[
                obis_references.ELECTRICITY_DELIVERED_TARIFF_1
            ].value
            msg_dsmr["ELECTRICITY_DELIVERED_TARIFF_2"] = telegram[
                obis_references.ELECTRICITY_DELIVERED_TARIFF_2
            ].value
            msg_dsmr["energy_delivered_net"] = (
                msg_dsmr["ELECTRICITY_DELIVERED_TARIFF_1"]
                + msg_dsmr["ELECTRICITY_DELIVERED_TARIFF_2"]
            )
            tarif = telegram[obis_references.ELECTRICITY_ACTIVE_TARIFF].value
            if tarif in self.tconfig["tarif"]:
                msg_dsmr["ELECTRICITY_ACTIVE_TARIFF"] = self.tconfig["tarif"][tarif]
            else:
                msg_dsmr["ELECTRICITY_ACTIVE_TARIFF"] = tarif

            if self.tconfig[terminal]["dsmr_version"] in ["5B", "5", "4"]:
                msg_dsmr["LONG_POWER_FAILURE_COUNT"] = telegram[
                    obis_references.LONG_POWER_FAILURE_COUNT
                ].value
                msg_dsmr["SHORT_POWER_FAILURE_COUNT"] = telegram[
                    obis_references.SHORT_POWER_FAILURE_COUNT
                ].value

                for fase in [1, 2, 3]:
                    self._proces_power_current_fase(fase, msg_dsmr, telegram)
                net_voltage_max = Decimal("0")
                if "INSTANTANEOUS_VOLTAGE_L1" in msg_dsmr:
                    net_voltage_max = max(
                        [msg_dsmr["INSTANTANEOUS_VOLTAGE_L1"], net_voltage_max]
                    )
                if "INSTANTANEOUS_VOLTAGE_L2" in msg_dsmr:
                    net_voltage_max = max(
                        [msg_dsmr["INSTANTANEOUS_VOLTAGE_L2"], net_voltage_max]
                    )
                if "INSTANTANEOUS_VOLTAGE_L3" in msg_dsmr:
                    net_voltage_max = max(
                        [msg_dsmr["INSTANTANEOUS_VOLTAGE_L3"], net_voltage_max]
                    )
                if net_voltage_max:
                    msg_dsmr["net_voltage_max"] = net_voltage_max

            # msg_dsmr['current_net_power'] = Decimal('0')
            # for fase in [1,2,3]:
            #    if f'current_net_power_l{fase}' in msg_dsmr:
            #        msg_dsmr['current_net_power'] += msg_dsmr[f'current_net_power_l{fase}']
            # if not msg_dsmr['current_net_power']:
            # Use global power indicators
            msg_dsmr["current_net_power"] = (
                (
                    msg_dsmr["CURRENT_ELECTRICITY_USAGE"]
                    - msg_dsmr["CURRENT_ELECTRICITY_DELIVERY"]
                )
                * Decimal("1000")
            ).quantize(Decimal("1."))

        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"DSMR current calculation for terminal '{terminal}' failed. "
                f"Check your config: Error: {e.args}",
            )

    def _process_gas(self, msg_dsmr, telegram):
        terminal = threading.currentThread().getName()
        if not self.tconfig[terminal]["gas_meter"]:
            # Skip gas meter
            return
        try:
            if self.tconfig[terminal]["dsmr_version"] in ["4", "5"]:
                G = telegram[obis_references.HOURLY_GAS_METER_READING]
                msg_dsmr["gas_consumption_total"] = G.values[1]["value"]
                msg_dsmr["timestamp_gas"] = datetime.timestamp(G.values[0]["value"])
            elif self.tconfig[terminal]["dsmr_version"] == "5B":
                G = telegram[obis_references.BELGIUM_HOURLY_GAS_METER_READING]
                msg_dsmr["gas_consumption_total"] = G.values[1]["value"]
                msg_dsmr["timestamp_gas"] = datetime.timestamp(G.values[0]["value"])
            elif self.tconfig[terminal]["dsmr_version"] == "2.2":
                G = telegram[obis_references.GAS_METER_READING]
                msg_dsmr["gas_consumption_total"] = G.values[6]["value"]
                msg_dsmr["timestamp_gas"] = datetime.timestamp(G.values[0]["value"])
            msg_dsmr["EQUIPMENT_IDENTIFIER_GAS"] = telegram[
                obis_references.EQUIPMENT_IDENTIFIER_GAS
            ].value
            # Set last value to the cache
            if not self.last_gas_update[terminal][0]:
                self.last_gas_update[terminal][0] = msg_dsmr["timestamp_gas"]
                self.last_gas_update[terminal][1] = msg_dsmr["gas_consumption_total"]
            # Calculate gas consumption / hour self.last_gas_update[terminal] = [0, Decimal('0')]
            if msg_dsmr["timestamp_gas"] > self.last_gas_update[terminal][0]:
                msg_dsmr["gas_consumption_hour"] = (
                    (
                        msg_dsmr["gas_consumption_total"]
                        - self.last_gas_update[terminal][1]
                    )
                    * Decimal("3600")
                    / Decimal(
                        msg_dsmr["timestamp_gas"] - self.last_gas_update[terminal][0]
                    )
                ).quantize(Decimal("0.001"))
                self.last_gas_update[terminal][0] = msg_dsmr["timestamp_gas"]
                self.last_gas_update[terminal][1] = msg_dsmr["gas_consumption_total"]
                self.last_gas_update[terminal][2] = msg_dsmr["gas_consumption_hour"]
            else:
                msg_dsmr["gas_consumption_hour"] = self.last_gas_update[terminal][2]
        except Exception as e:
            hybridlogger.ha_log(
                self.logger,
                self.hass_api,
                "WARNING",
                f"DSMR gas_meter reading for terminal '{terminal}' failed. "
                f"Check your config: Error: {e.args}",
            )

    def dsmr_serial_callback(self, telegram):
        terminal = threading.currentThread().getName()
        """Callback that prints telegram values."""
        if obis_references.P1_MESSAGE_TIMESTAMP in telegram:
            # Time stamp from Telegram
            TS = datetime.timestamp(
                telegram[obis_references.P1_MESSAGE_TIMESTAMP].value
            )
        else:
            # Create time stamp
            TS = int(datetime.timestamp(datetime.now()))
        # Use power value to sync DSMR 5 telegram flod. DSMR 4 and 2 log possible every 10 sec
        P1 = telegram[obis_references.CURRENT_ELECTRICITY_USAGE]
        P2 = telegram[obis_references.CURRENT_ELECTRICITY_DELIVERY]
        P = ((P2.value - P1.value) * Decimal("1000")).quantize(Decimal("1."))

        # Sync telegrams if there is a power change of after 10s
        if (
            (self.cache[terminal] != P)
            or (TS - self.ts_last_telegram[terminal] > 9.5)
            or (self.sync[terminal] == 10)
        ):
            self.cache[terminal] = P
            self.sync[terminal] = 0
            self.ts_last_telegram[terminal] = TS
        # Process telegram if synced
        if self.sync[terminal] == 0:
            # create uniform dsmr message structure for futher processinh
            msg_dsmr = {}
            # Add generic values
            msg_dsmr["timestamp"] = TS
            msg_dsmr["EQUIPMENT_IDENTIFIER"] = telegram[
                obis_references.EQUIPMENT_IDENTIFIER
            ].value
            msg_dsmr["terminal"] = terminal
            msg_dsmr["plant_id"] = self.tconfig[terminal]["plant_id"]
            msg_dsmr["total_energy_offset"] = self.tconfig[terminal][
                "total_energy_offset"
            ]
            # msg_dsmr['advertize_power_fields'] =  self.tconfig[terminal]['advertize_power_fields']
            # msg_dsmr['advertize_dsrm_solar_fields'] =  self.tconfig[terminal]['advertize_dsrm_solar_fields']

            # Read and process more accurate power calculations DSMR 5
            self._process_power_details(msg_dsmr, telegram)

            # Process Gas
            self._process_gas(msg_dsmr, telegram)

            # Send back data to data logger
            self.dsmr_callback(terminal, msg_dsmr)

        # Increase terminal sync counter
        self.sync[terminal] += 1
